from typing import Any

from fastapi import APIRouter, Depends

from music.context import context
from music.models import Artist
from music.models.response_types import ArtistInfo
from music.server.servers.artists import Server

app = APIRouter(prefix="/artists")
server = Server()


def parse_limit(limit: str | None = None) -> int | None:
    if limit is None:
        parsed_limit = 20
    elif not limit:
        parsed_limit = None
    else:
        parsed_limit = int(limit)
    return parsed_limit


@app.get("")
async def load_saved_artists(
    offset: int = 0,
    limit: int | None = Depends(parse_limit),
) -> list[dict[str, Any]]:
    artists = server.load_saved_artists(offset=offset, limit=limit)
    return list(artists)


@app.get("/search")
async def search_artist(name: str) -> list[dict[str, Any]]:
    artists = server.create_artist_search_results(name)
    return list(artists)


@app.get("/add")
async def save_new_artist(id_: str, name: str) -> None:
    artist = Artist(id_, name)
    return context.storage.save_new_artist(artist)


@app.get("/toggle")
async def toggle_artist_type(id_: str) -> None:
    artists: list[Artist] = context.storage.artists
    for artist in artists:
        if artist.id == id_:
            artist.toggle_type()
    context.storage.artists = artists


@app.get("/recommendations")
async def recommend_artists() -> list[ArtistInfo]:
    return server.create_artist_recommendations()



from collections.abc import Iterable
from typing import TypeVar

import cli

T = TypeVar("T")


def track_progress(
    sequence: Iterable[T],
    description: str = "",
    unit: str = "item",
    total: int | None = None,
    *,
    cleanup_after_finish: bool = True,
) -> Iterable[T]:
    yield from cli.track_progress(
        sequence,
        description,
        unit,
        total,
        cleanup_after_finish=cleanup_after_finish,
    )



import os
from contextlib import AbstractContextManager
from functools import cached_property

from package_utils.context import Context as Context_

from music.clients import spotify
from music.models import Config, Options, Secrets
from music.storage.storage import Storage


class Context(Context_[Options, Config, Secrets]):
    @cached_property
    def spotify_client(self) -> spotify.Client:
        return spotify.Client(self.secrets)

    @cached_property
    def storage(self) -> Storage:
        return Storage()

    @cached_property
    def is_running_in_ci(self) -> bool:
        return (
            "GITHUB_ACTIONS" in os.environ and "PYTEST_CURRENT_TEST" not in os.environ
        )

    @cached_property
    def ci_context(self) -> AbstractContextManager[None] | None:
        ci_context = None
        if context.is_running_in_ci:  # pragma: nocover
            # imports optional dependencies
            try:
                from music.utils.ci_context import CIContext
            except ImportError:
                pass
            else:
                ci_context = CIContext()
        return ci_context


context = Context(Options, Config, Secrets)



from dataclasses import dataclass

from music.context import context
from music.models import Artist
from music.models.response_types import Album
from music.storage.artist import Storage


@dataclass
class ArtistUpdater:
    artist: Artist

    def __post_init__(self) -> None:
        self.storage = Storage(self.artist)

    def check_for_new_songs(self) -> None:
        if self.artist.is_normal:
            self.check_top_songs()
        else:
            self.check_all_songs()

    def check_top_songs(self) -> None:
        tracks = context.spotify_client.top_songs(self.artist.id)
        new_tracks = [
            track for track in tracks if track.id not in self.storage.top_tracks
        ]
        context.storage.save_new_tracks(new_tracks)
        new_tracks_dict = {track.id: track.name for track in tracks}
        if new_tracks_dict:
            self.storage.top_tracks |= new_tracks_dict

    def check_all_songs(self) -> None:
        song_count = context.spotify_client.album_count(self.artist.id)
        if song_count > self.storage.albums_count:
            album_types = ("album", "single")
            for album_type in album_types:
                self.check_albums(album_type)
            self.storage.albums_count = song_count

    def check_albums(self, album_type: str) -> None:
        current_amount = context.spotify_client.album_count(
            self.artist.id,
            album_type=album_type,
        )
        saved_amount = self.storage.get_album_count(album_type)
        added_amount = current_amount - saved_amount
        if added_amount > 0:
            self.save_new_albums(album_type, added_amount)

    def save_new_albums(self, album_type: str, amount: int) -> None:
        new_albums = context.spotify_client.albums(
            self.artist.id,
            album_type=album_type,
            amount=amount,
        )
        for album in new_albums:
            self.save_new_album(album)
        num_new_albums = len(new_albums)
        self.storage.add_album_count(album_type, num_new_albums)

    def save_new_album(self, album: Album) -> None:
        tracks = context.spotify_client.album_songs(album)
        track_ids = [track.id for track in tracks]
        # popularity and release_date needed
        tracks_info = context.spotify_client.songs(track_ids)
        context.storage.save_new_tracks(tracks_info)
        album_tracks = {track.id: track.name for track in tracks_info}
        self.storage.albums |= {album.id: album_tracks}



from dataclasses import dataclass

import cli
from spotdl.console.download import download

from music.context import context
from music.models import Path

from . import downloaded_songs_processor, spotdl


@dataclass
class Downloader:
    track_ids: list[str]

    def start(self) -> None:
        self.clear_downloads_path()
        self.download_songs()
        with cli.status("Processing downloads"):
            downloaded_songs_processor.run()
        self.update_downloaded_tracks()
        self.update_tracks_to_download()

    @classmethod
    def clear_downloads_path(cls) -> None:
        # existing downloads raise errors and can be removed because
        # successfully downloaded songs have already been moved to other folder
        for path in Path.downloaded_songs.iterdir():
            path.unlink()

    def create_download_urls(self) -> list[str]:
        return [
            id_ if id_.startswith("http") else f"https://open.spotify.com/track/{id_}"
            for id_ in self.track_ids
        ]

    def download_songs(self) -> None:
        tries_left = context.config.download_retries
        urls = self.create_download_urls()
        while urls and tries_left:
            download(urls, spotdl.downloader)
            urls = list(Path.downloaded_songs.glob("*.spotdlTrackingFile"))
            tries_left -= 1

        if urls and not tries_left:
            message = "Max download retries reached"
            raise RuntimeError(message)

    def update_downloaded_tracks(self) -> None:
        tracks_to_download = context.storage.tracks_to_download
        downloaded_tracks = {id_: tracks_to_download[id_] for id_ in self.track_ids}
        context.storage.downloaded_tracks |= downloaded_tracks

    def update_tracks_to_download(self) -> None:
        tracks_to_download = context.storage.tracks_to_download
        for id_ in self.track_ids:
            tracks_to_download.pop(id_)
        context.storage.tracks_to_download = tracks_to_download



from dataclasses import dataclass


@dataclass
class Spotify:
    client_id: str
    client_secret: str


@dataclass
class Secrets:
    genius_token: str
    spotify: Spotify
    phone_connection: str



from typing import Any

from fastapi import APIRouter

from music.context import context
from music.models.response_types import Track
from music.server.servers.songs import Server

app = APIRouter(prefix="/songs")
server = Server()


@app.get("/search")
async def search_song(name: str) -> list[dict[str, Any]]:
    songs = server.create_song_search_results(name)
    return list(songs)


@app.get("/add")
async def save_new_track(id_: str) -> None:
    songs = context.spotify_client.songs([id_])
    context.storage.save_new_tracks(songs)


@app.get("/recommendations")
async def recommend_songs() -> list[Track]:
    songs = server.create_song_recommendations()
    return list(songs)



import sys

import cli

from music.context import context
from music.utils.progress import track_progress

from .artist import ArtistUpdater


def check_for_new_songs() -> None:
    description = "Checking for new songs"
    cli.console.clear_live()
    if context.is_running_in_ci:
        check_for_new_songs_with_print_progress(description)  # pragma: nocover
    else:
        check_for_new_songs_with_rich_progress(description)


def check_for_new_songs_with_print_progress(
    description: str,
) -> None:  # pragma: nocover
    cli.console.print(description)
    number_of_artists = len(context.storage.artists)
    for i, artist in enumerate(context.storage.artists):
        cli.console.print(f"Checking {artist.name} ({i +1}/{number_of_artists})")
        sys.stdout.flush()
        ArtistUpdater(artist).check_for_new_songs()


def check_for_new_songs_with_rich_progress(description: str) -> None:
    artists = track_progress(
        context.storage.artists,
        description=description,
        unit="artists",
    )
    for artist in artists:
        ArtistUpdater(artist).check_for_new_songs()



from dataclasses import dataclass

from .path import Path


@dataclass
class Config:
    secrets_path = Path.secrets
    number_of_recommendations: int = 20
    max_new_recommendation_tries: int = 10
    download_chunk_size: int = 20
    download_retries: int = 5
    phone_upload_port: int = 2222



from types import TracebackType
from typing import Any, TypeVar, cast

import superpathlib
from simple_classproperty import classproperty

T = TypeVar("T", bound="Path")


class Path(superpathlib.Path):
    @property  # type: ignore[override]
    def yaml(self) -> dict[str, Any] | int:
        # cache results in json
        if self.json_path.mtime < self.mtime:
            self.json_path.json = super().yaml
        return self.json_path.json  # type: ignore[return-value]

    @yaml.setter
    def yaml(self, value: dict[str, Any] | int) -> None:
        super(superpathlib.Path, superpathlib.Path(self)).__setattr__("yaml", value)
        self.json_path.json = value  # type: ignore[assignment]

    @property
    def json_path(self: T) -> T:
        return self.with_suffix(".json")

    def __exit__(
        self,
        exception_type: type[BaseException] | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        super().__exit__(exception_type, exception_value, traceback)
        super(Path, self.json_path).__exit__(exception_type, exception_value, traceback)

    @classmethod
    @classproperty
    def source_root(cls: type[T]) -> T:
        return cls(__file__).parent.parent

    @classmethod
    @classproperty
    def assets(cls: type[T]) -> T:
        path = cls.script_assets / cls.source_root.name
        return cast(T, path)

    @classmethod
    @classproperty
    def config(cls: type[T]) -> T:
        path = cls.assets / "config" / "config.yaml"
        return cast(T, path)

    @classmethod
    @classproperty
    def download_info(cls: type[T]) -> T:
        path = cls.assets / "downloads"
        return cast(T, path)

    @classmethod
    @classproperty
    def download_ids(cls: type[T]) -> T:
        path = cls.download_info / "ids.yaml"
        return cast(T, path)

    @classmethod
    @classproperty
    def fails(cls: type[T]) -> T:
        path = cls.download_info / "fails.yaml"
        return cast(T, path)

    @classmethod
    @classproperty
    def to_download(cls: type[T]) -> T:
        path = cls.assets / "cache" / "to_download.yaml"
        return cast(T, path)

    @classmethod
    @classproperty
    def artist_assets(cls: type[T]) -> T:
        path = cls.assets / "artists"
        return cast(T, path)

    @classmethod
    @classproperty
    def artists(cls: type[T]) -> T:
        path = cls.artist_assets / "artists.yaml"
        return cast(T, path)

    @classmethod
    @classproperty
    def recommendations(cls: type[T]) -> T:
        path = cls.artist_assets / "recommendations.yaml"
        return cast(T, path)

    @classmethod
    @classproperty
    def secrets(cls: type[T]) -> T:
        # don't use json caching here
        path = superpathlib.Path(cls.assets) / "tokens" / "tokens"
        return cast(T, path)

    @classmethod
    @classproperty
    def cache_assets(cls: type[T]) -> T:
        path = cls.assets / "cache"
        return cast(T, path)

    @classmethod
    @classproperty
    def cache(cls: type[T]) -> T:
        path = cls.cache_assets / ".cache"
        return cast(T, path)

    @classmethod
    @classproperty
    def download_assets(cls: type[T]) -> T:
        return cls("/") / "media" / "backup" / "Music"

    @classmethod
    @classproperty
    def downloaded_songs(cls: type[T]) -> T:
        path = cls.download_assets / "downloads"
        return cast(T, path)

    @classmethod
    @classproperty
    def processed_songs(cls: type[T]) -> T:
        path = cls.download_assets / "processed"
        return cast(T, path)

    @classmethod
    @classproperty
    def all_songs(cls: type[T]) -> T:
        path = cls.download_assets / "all"
        return cast(T, path)

    @classmethod
    @classproperty
    def deleted(cls: type[T]) -> T:
        path = cls.download_assets / "deleted"
        return cast(T, path)

    phone = "Music"



from __future__ import annotations

import typing
from typing import TYPE_CHECKING

from package_utils.storage import CachedFileContent

from .artist_paths import Paths

if TYPE_CHECKING:
    from music.models import Artist  # pragma: nocover


class StorageModel:
    albums: dict[str, dict[str, str]]
    albums_count: int
    album_counts: dict[str, int]
    top_tracks: dict[str, str]

    def __init__(self, artist: Artist) -> None:
        self.artist = artist

    def get_album_count(self, album_type: str) -> int:
        return self.album_counts.get(album_type, 0)

    def set_album_count(self, album_type: str, value: int) -> None:
        self.album_counts[album_type] = value

    def add_album_count(self, album_type: str, value: int) -> None:
        new_count = self.get_album_count(album_type) + value
        self.set_album_count(album_type, new_count)


class Storage(StorageModel):
    def __new__(cls, artist: Artist) -> Storage:  # noqa: PYI034
        paths = Paths(artist)

        class ImplementedStorage(StorageModel):
            albums = paths.albums.cached_content  # type: ignore[assignment]
            albums_count = CachedFileContent(paths.albums_count, default=0)  # type: ignore[assignment]
            album_counts = paths.album_counts.cached_content  # type: ignore[assignment]
            top_tracks = paths.top_tracks.cached_content

        storage = ImplementedStorage(artist)
        return typing.cast(Storage, storage)



import random
from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from typing import Any

from music.context import context
from music.models.response_types import Track


@dataclass
class Server:
    number_of_song_recommendation_seeds: int = 5

    @classmethod
    def create_song_search_results(cls, name: str) -> Iterator[dict[str, Any]]:
        songs = context.spotify_client.search_song(name)
        for song, is_downloaded in cls.check_is_downloaded(songs):
            yield song.dict() | {"downloaded": is_downloaded}

    @classmethod
    def check_is_downloaded(
        cls,
        songs: Iterable[Track],
    ) -> Iterator[tuple[Track, bool]]:
        downloaded_tracks = context.storage.downloaded_tracks
        downloaded_track_names = set(downloaded_tracks.values())
        for song in songs:
            is_downloaded = (
                song.id in downloaded_tracks or song.full_name in downloaded_track_names
            )
            yield song, is_downloaded

    def create_song_recommendations(self) -> Iterator[Track]:
        downloaded_song_ids = list(context.storage.downloaded_tracks.keys())
        random.shuffle(downloaded_song_ids)
        seed_ids = downloaded_song_ids[: self.number_of_song_recommendation_seeds]
        recommendations = context.spotify_client.song_recommendations(seed_ids)
        for song, is_downloaded in self.check_is_downloaded(recommendations):
            if not is_downloaded:
                yield song



from collections.abc import Awaitable, Callable

import powertrace
from fastapi import FastAPI, Request, Response
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import PlainTextResponse

from .routers import artists, songs


def create_app() -> FastAPI:
    app = FastAPI()
    configure_cors(app)
    configure_exception_handler(app)
    configure_routes(app)
    return app


def configure_routes(app: FastAPI) -> None:
    routers = artists.app, songs.app
    for router in routers:
        app.include_router(router)


def configure_exception_handler(app: FastAPI) -> None:
    @app.middleware("http")
    async def exception_handling(
        request: Request,
        call_next: Callable[[Request], Awaitable[Response]],
    ) -> Response:
        try:
            response = await call_next(request)
        except Exception as exception:  # noqa: BLE001
            powertrace.visualize_traceback(exit_after=False, repeat=False)
            response = PlainTextResponse(str(exception), status_code=500)
        return response


def configure_cors(app: FastAPI) -> None:
    origins = ["https://music.com", "http://localhost:8080"]
    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,
        allow_credentials=True,
        allow_headers=["*"],
    )



import sys
from contextlib import AbstractContextManager
from types import TracebackType

import backup.models
import cli  # pragma: nocover
from backup.backup import Backup  # pragma: nocover
from backup.main.mount import Mounter  # pragma: nocover

from music.context import context  # pragma: nocover
from music.models import Path  # pragma: nocover


class CIContext(AbstractContextManager[None]):  # pragma: nocover
    def __init__(self) -> None:
        remote_home = backup.models.Path.remote / "home" / "quinten"
        dest = remote_home / Path.assets.relative_to(Path.HOME)
        self.backup = Backup(source=Path.assets, dest=dest)

    def __enter__(self) -> None:
        cli.console._force_terminal = True  # noqa: SLF001
        context.options.upload_to_phone = False
        remote = f"{backup.models.Path.remote}Music"
        Mounter(remote=remote, path=Path.download_assets).run()
        self.print("downloading configurations..")
        self.backup.capture_pull()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.print("uploading results..")
        self.backup.capture_push()

    @classmethod
    def print(cls, message: str) -> None:
        cli.console.print(message)
        sys.stdout.flush()



import os

from package_utils.context.entry_point import Context, create_entry_point

from music.context import context
from music.main import main
from music.models import Config, Options, Secrets


def configure_no_secret_file(context: Context[Options, Config, Secrets]) -> None:
    if "GITHUB_ACTIONS" in os.environ:
        context.config.secrets_path = None  # pragma: nocover


entry_point = create_entry_point(
    main,
    context,
    context_creation_callback=configure_no_secret_file,
)



from dataclasses import dataclass
from typing import Annotated

import typer

from .path import Path

clean_download_ids_help = (
    "Spotify changes the ids of its tracks from time to time.\n"
    "This command saves the most recent id of all downloaded tracks"
)

clean_download_ids_option = typer.Option(help=clean_download_ids_help)


@dataclass
class Options:
    config_path: Path = Path.config
    upload_to_phone: bool = True
    add: Annotated[str | None, typer.Option(help="song to add")] = None
    clean_download_ids: Annotated[bool, clean_download_ids_option] = False
    fix_mtimes_on_phone: bool = False



from music.models import Secrets
from music.models.response_types import (
    Album,
    Albums,
    AlbumTrack,
    AlbumTracks,
    ArtistInfo,
    Artists,
    ArtistSearch,
    RecommendedTracks,
    Track,
    Tracks,
    TrackSearch,
)

from . import _spotify
from .utils import combine_chunks, combine_offsets


class Client:
    def __init__(self, secrets: Secrets) -> None:
        self.client = _spotify.Spotify(secrets)

    def search_song(self, name: str) -> list[Track]:
        songs = self.client.search(name, type="track")
        return TrackSearch.from_dict(songs["tracks"]).items

    def search_artist(self, name: str) -> list[ArtistInfo]:
        artists = self.client.search(name, type="artist")
        return ArtistSearch.from_dict(artists["artists"]).items

    @combine_chunks
    def artists(self, ids: list[str]) -> list[ArtistInfo]:
        artists = self.client.artists(ids)
        return Artists.from_dict(artists).artists

    def related_artists(self, id_: str) -> list[ArtistInfo]:
        artists = self.client.artist_related_artists(id_)
        return Artists.from_dict(artists).artists

    def song_recommendations(self, track_ids: list[str]) -> list[Track]:
        songs = self.client.recommendations(seed_tracks=track_ids)
        return RecommendedTracks.from_dict(songs).tracks

    def top_songs(self, id_: str) -> list[Track]:
        songs = self.client.artist_top_tracks(id_)
        return Tracks.from_dict(songs).tracks

    @combine_chunks
    def songs(self, ids: list[str]) -> list[Track]:
        songs = self.client.tracks(ids)
        return Tracks.from_dict(songs).tracks

    def albums(self, artist_id: str, album_type: str, amount: int) -> list[Album]:
        return self._albums(amount, artist_id, album_type)

    @combine_offsets
    def _albums(
        self,
        artist_id: str,
        album_type: str,
        limit: int,
        offset: int,
    ) -> list[Album]:
        albums = self.client.artist_albums(
            artist_id,
            limit=limit,
            offset=offset,
            album_type=album_type,
        )
        return Albums.from_dict(albums).items

    def album_songs(self, album: Album) -> list[AlbumTrack]:
        return self._album_songs(album.total_tracks, album.id)

    @combine_offsets
    def _album_songs(self, id_: str, offset: int, limit: int) -> list[AlbumTrack]:
        songs = self.client.album_tracks(id_, limit=limit, offset=offset)
        return AlbumTracks.from_dict(songs).items

    def album_count(self, artist_id: str, album_type: str = "album,single") -> int:
        albums = self.client.artist_albums(artist_id, limit=1, album_type=album_type)
        return Albums.from_dict(albums).total
